/* 
 * (C) 2003 Ulrich Hecht <uli@emulinks.de>
 * if2a patch by D. Gauchard <deyv@free.fr>
 * F2A Ultra support by Vincent Rubiolo <vincent.rubiolo@free.fr>
 * Licensed under the terms of the GNU Public License version 2
 */
 
// Cart/ROM functions

#include "f2acarts.h"
#include "f2aultra.h"
#include "f2autils.h"
#include "f2aio.h"

int f2a_reduce_forced = 0;
int f2a_reduce_allowed = 1;
int f2a_cartsize_mb = -1;		/* autodetect, see f2a_get_type() */
int f2a_write_block_size_log2 = -1;	/* autodetect, see f2a_get_type() */
int f2a_rom_block_size_log2 = -1;	/* autodetect, see f2a_get_type() */
int f2a_thorough_comp = 0;
int f2a_use_loader = 1;
int f2a_correct_header_allowed = 1;
int f2a_force_burn = 0;

int select_loader (enum cart_type_e cart_type, char* loader_file)
{
	int result = -1;

	if (loader.size == 0)
	{
		switch (cart_type)
		{
			case CART_TYPE_PRO:
				result = select_loader_pro(loader_file);
				break;
			case CART_TYPE_ULTRA:
				result = select_loader_ultra(loader_file);
				break;
			case CART_TYPE_TURBO:
				printerr("looking for TURBO cart testers\n");
				result = select_loader_pro(loader_file);
				break;
			case CART_TYPE_UNDEF:
				printerr("Undefined cart type, please use -t\n");
				result = -1;
				break;
		}
		return result;
	}
	else
		return 0;
}

int has_good_header (unsigned char* rom)
{
	/* 
	 * FIXME
	 * Weak implementation : no check on required checksum fields
	 */
	return memcmp(loader.data+4, rom+4, HEADERSIZE_B)==0? 1: -1;
}

void correct_header (unsigned char* rom, unsigned char* name)
{
	unsigned char romname[13];
	unsigned char romcode[5];
	unsigned char* shortname;
	
	memset(romname, 32, 12);
	memset(romcode, 32, 12);
	romname[12] = romcode[4] = 0;
	
	for (shortname = name + strlen(name) - 1;
	     shortname > name && (*(shortname - 1)) != '/' && (*(shortname - 1)) != '\\';
	     shortname--);
	
	memcpy(romname, shortname, MIN(strlen(shortname), 12));
	memcpy(romcode, shortname, MIN(strlen(shortname), 4));
	
	if (!isalnum(rom[0xa0]))
		memcpy(&rom[0xa0], romname, 12);
	if (!isalnum(rom[0xac]))
		memcpy(&rom[0xac], romcode, 4);
	//if (!isalnum(rom[0xb0]))
	//	rom[0xb0] = '0';
	
	if (has_good_header(rom) < 0)
	{
		printerr("No standard header found in '%s', rewriting it\n", romname);
		memcpy(rom, loader.data, HEADERSIZE_B);
	}
}

void display_map (unsigned char* rom)
		{
	int i;
	
	print("Game Title:	");
	for (i = 0xa0; i < 0xac; i++)
		print("%c", ASCII(rom[i]));
	print("\nGame Code:	AGB-");
	for (i = 0xac; i < 0xb0; i++)
		print("%c", ASCII(rom[i]));
	print("\nMaker Code:	");
	for (i = 0xb0; i < 0xb2; i++)
		print("%c", ASCII(rom[i]));
	print("\n");
	//print("\n0x96:		0x%02x\n", rom[0xb2]);
}

unsigned char* romname (unsigned char* rom)
{
	int len;
	static char name [13];
	
	for (len = 0; len < 12; len++)
		name[len] = ASCII(rom[0xa0 + len]);
	for (len--; len>=0 && (name[len]==32 || name[len]=='.'); len--);
	name[len+1] = 0;
	return name;
}

int display_cart_map (void)
{
	int i;
	unsigned char rom[LOADUNIT];
	
	if (f2a_sim > 1)
		return 0;

	for (i = GBA_ROM; i < GBA_ROM + f2a_max_rom_size; i += f2a_rom_block_size)
	{
		print("Trying address 0x%x...\r", i);
		if (f2a_readmem(rom, i, LOADUNIT) == -1)
			return -1;
		if (has_good_header(rom) > -1)
		{
			print("\n");
			display_map(rom);
		}
	}
	print("\n");

	return 0;
}

void auto_readandsave_rom (int numfiles, char** files)
{
	int i;
	int indexfile;
	int nextblockloaded;
	char* name;
	char autoname[64];
	unsigned char rom[f2a_rom_block_size];
	
	assert(f2a_rom_block_size >= LOADUNIT);
	
	if (f2a_sim > 1)
		return;

	indexfile = 0;
	i = GBA_ROM;
	f2a_readmem(rom, i, LOADUNIT);
	do
	{
		nextblockloaded = 0;
		
		print("Trying address 0x%x...\r", i);
		if (has_good_header(rom) != -1)
		{
			/* Prepare filename */
			name = romname(rom);
			strcpy(autoname, GBA_HEADNAME);
			strcat(autoname, name);
			strcat(autoname, GBA_EXTNAME);
			
			print("\n");
			print("Detected rom '%s'\n", name);
			
			/* If no arg is given, autoname; otherwise, use specified name */
			if (indexfile >= numfiles)
				name = autoname;
			else
				name = files[indexfile];
			indexfile++;
			
			if (strcmp(name, "@") == 0)
				name = autoname;
			
			if (strcmp(name, ".") == 0) /* user don't want this rom */
			{
				if (f2a_verbose)
					print("Skipping on user request\n");
			}
			else
			{
				FILE* f;
				if (f2a_verbose)
					print("Using name file name '%s'\n", name);
				if ((f = fopen(name, "wb")) == NULL)
					printerrno("fopen(%s)", name);
				else
				{
					if (f2a_readmem(rom, i, f2a_rom_block_size) < 0)
						return;
					do /* read data up to next ROM/valid header */
					{
						if (fwrite(rom, f2a_rom_block_size, 1, f) != 1)
						{
							if (ferror(f))
								printerrno("write(%s)", name);
							else
								printerr("could not write to file %s\n", name);
							break;
						}
						if ((i+=f2a_rom_block_size) < GBA_ROM + f2a_max_rom_size && f2a_readmem(rom, i, f2a_rom_block_size) < 0)
							return;
					} while (i < GBA_ROM + f2a_max_rom_size && has_good_header(rom) == -1);
					fclose(f);
					nextblockloaded = 1;
				}
			}
		}
		
		if (!nextblockloaded) // continue to look for ROM/valid header
		{
			if ((i+=f2a_rom_block_size) < GBA_ROM + f2a_max_rom_size && f2a_readmem(rom, i, LOADUNIT) < 0)
				return;
		}
	} while (i < GBA_ROM + f2a_max_rom_size);
	
	print("\n");
}

void display_memory_map (unsigned char* rom, int size)
{
	int i;

	for (i = 0; i < size; i += f2a_rom_block_size)
	{
		print("Trying address 0x%x...\r", i);
		if (has_good_header(&rom[i]) > -1)
		{
			print("\n");
			display_map(&rom[i]);
		}
	}
}

int is_same (int offset, int size, unsigned char* mem)
{
	int ret, i;
	unsigned char rom [LOADUNIT];
	
	if (f2a_sim > 1)
		return -1;

	for (i = offset; i < offset + size; i += LOADUNIT)
	{
		// check beginning
		f2a_readmem(rom, GBA_ROM + i, LOADUNIT);
		ret = memcmp(rom, &mem[i], LOADUNIT);
		if (f2a_verbose > 1 && ret != 0)
			print("(offset 0x%x-0x%x BAD)\n", i, i + LOADUNIT - 1);
		if (ret != 0)
			return -1;
	}
	return 1;
}

int has_same_data (int offset, int size, unsigned char* mem)
{
	if (f2a_thorough_comp)
	{
		int i;
		for (i = offset; i < offset + size; i += f2a_rom_block_size)
			if (is_same(i, f2a_rom_block_size, mem) == -1)
				return -1;
	}
	else
	{
		// check beginnning
		if (is_same(offset, LOADUNIT, mem) == -1)
			return -1;

		// roughly check end
		size &= ~(LOADUNIT-1);
		if (is_same(offset + size - LOADUNIT, LOADUNIT, mem) == -1)
			return -1;
	}
	return 1;
}

int convsize (char* size)
{
	char* unit = size;
	while (isdigit(*unit))
		unit++;
	if (strlen(unit) == 2)
	{
		int byte;
		switch (tolower(unit[0]))
		{
		case 'k':	byte = 1<<10; break;
		case 'm':	byte = 1<<20; break;
		default:	byte = -8;
		}
		switch (unit[1])
		{
		case 'b':	byte >>= 3; break;
		case 'B':	break;
		default:	byte = -1;
		}
		if (byte > 0)
		{
			*unit = 0; // THIS IS BAD, we should use a const char* as argument
			return atoi(size) * byte;
		}
	}
	printerr("size '%s' not recognized\n", size);
	f2a_exit(1);
	return -1;
}

int convf2ablock (char* block, int* offset, int* size )
{
	int len = strlen(block);

	if (len == 0 || len > 3)
		return -1;

	*offset = (block[0] - '1') << 16;
	if (block[1])
	{
		*offset += (tolower(block[1]) - 'a') << 15;
		
		if (block[2])
		{
			*offset += (block[2] - '1') << 13;
			*size = 1 << 13;
		}
		else
			*size = 1 << 15;
	}
	else
		*size = 1 << 16;
	if (f2a_verbose)
		print("Block %s is SRAM is (offset=0x%x/%ibytes size=0x%x/%ibytes)\n",
			block, *offset, *offset, *size, *size);
	return 0;
}

//////////////////////////////////////////////////////////////////////////
// check whole size

#define STAT							\
		if (index == -1)				\
			st.st_size = loader.size;		\
		else if (stat(files[index], &st) == -1)		\
		{						\
			perror(files[index]);			\
			return -1;				\
		} 

int get_wholesize (int numfiles, char** files)
{
	int index;
	struct stat st;
	int wholesize = 0;
	int roundedfilesize;
	
	// stat every file
	for (index = f2a_use_loader? -1: 0; index < numfiles; index++)
	{
		if (f2a_verbose)
			print("Stat file: %s\n", index==-1? "Loader": files[index]);

		STAT

		roundedfilesize = st.st_size;
		adjust_rom_size(&roundedfilesize); // round on f2a_rom_block_size
		wholesize += roundedfilesize;
	}
	
/*
	if (wholesize > f2a_max_rom_size + loader.size)
	{
		printerr("You have not enough memory !\n");
		f2a_exit(1);
	}
*/
	if (wholesize > f2a_max_rom_size)
		// We hope that we will be able to reduce overall content by loader.size
		wholesize = f2a_max_rom_size;
	
	adjust_write_size(&wholesize); // round on f2a_write_block_size
	return wholesize;
}

//////////////////////////////////////////////////////////////////////////
// this one is simple, it writes one file to sram

int f2apro_loadandwrite_sram (char* file, int offset, int size)
{
	int ret;
	char* sram;
	
	if ((sram = prepare_loadandwrite_sram (file, offset, size)) == NULL)
		return -1;
	ret = f2a_writemem(sram, GBA_SRAM, offset, size, size);
	free(sram);
	return ret;
}

char* prepare_loadandwrite_sram (char* file, int offset, int size)
{
	FILE* sramf;
	struct stat st;
	char* sram = NULL;
	
	if (stat(file, &st) == -1)
	{
		perror(file);
		return NULL;
	}

	// some checks
	if (st.st_size > size)
	{
		printerr("File too big (0x%x bytes) to fit in 0x%x bytes\n",
			(int)st.st_size, size);
		return NULL;
	}

	// allocate memory
	if ((sram = (unsigned char*)malloc(size)) == NULL)
	{
		printerr("Cannot allocate 0x%x/%i bytes for sram management - ", size, size);
		perror("malloc");
		return NULL;
	}
	memset(sram, 0xff, size);

	// load files into "allocated rom"
	if (f2a_verbose)
		print("Loading file: %s (size=0x%x)\n", file, (int)st.st_size);

	if ((sramf = fopen(file, "rb")) == NULL)
	{
		printerrno("fopen(%s)", file);
		free(sram);
		return NULL;
	}
	if (fread(sram, size, 1, sramf) != 1)
	{
		if (ferror(sramf))
			printerrno("fread(%s)", file);
		else
			printerr("failed to read from file %s", file);
		free(sram);
		return NULL;
	}
	fclose(sramf);
	
	return sram;
}


//////////////////////////////////////////////////////////////////////////
// automatic check cart and burn if necessary

int auto_loadandburn_rom (enum cart_type_e cart_type, int clean_rom, int numfiles, char** files)
{
	typedef struct
	{
		int	offset;
		int	size;
	} chunk_s;

	int			burnstart		= -1;
	int			wholesize		= get_wholesize(numfiles, files);
	rom_map_locator_s*	locator			= NULL;
	rom_map_s*		rom_map			= NULL;
	int			rom_map_entries		= 0;
	int			chunks_number		= numfiles + 1;
	int			chunks_used		= 0;
	int			whole_loader_size	= 0;
	int			reduce			= 0;

	int			index;
	FILE*			ROMf;
	struct stat		st;
	int			loadedsize;
	int			roundedfilesize;
	char*			image;				// image file that will contain all cart contents
	chunk_s			chunks [chunks_number];		// chunks to burn array descriptor
	content_desc		desc;
	
	if (!wholesize)
		return -1;
	
	// allocate image memory
	if ((image = (unsigned char*)malloc(wholesize)) == NULL)
	{
		printerrno("Cannot allocate 0x%x/%i bytes for rom management - malloc", wholesize, wholesize);
		return -1;
	}
	memset(image, 0xff, wholesize);
	
	// load
	loadedsize = 0;
	for (index = (f2a_use_loader? -1: 0); index < numfiles; index++)
	{
		print("\n");
		
		STAT
		
		if (index == -1) // manage loader, fit rom map
		{
			int i, real_loader_size;
			char last = loader.data[loader.size - 1];

			assert(loader.size > 0);

			roundedfilesize = loader.size;
			adjust_rom_size(&roundedfilesize);
			// copy f2aloader into image
			memcpy(image, loader.data, loader.size);
			// pad chunk with last loader byte
			memset(image + loader.size, last, roundedfilesize - loader.size);

			// try to find a location for the rom map
			for (i = roundedfilesize - 2; image[i] == last; i--);
			real_loader_size = i + 2;
			assert(real_loader_size <= loader.size);
			
			// check whether rom map can fit 
			do
			{
				int spare_size = roundedfilesize - real_loader_size;
				rom_map_entries = spare_size > sizeof(rom_map_locator_s)? (spare_size - sizeof(rom_map_locator_s)) / sizeof(rom_map_s): 0;
				if (f2a_verbose)
					print("Spare size for rom map at rounded loader's end is %i bytes, which gives room for %i roms.\n", spare_size, rom_map_entries);
				if (rom_map_entries >= MAP_MINIMUM_ENTRIES)
				{
					int rom_map_locator = roundedfilesize - sizeof(rom_map_locator_s);
					locator = (rom_map_locator_s*)&image[rom_map_locator];
					locator->absolute_location = hton32(roundedfilesize - sizeof(rom_map_locator_s) - rom_map_entries * sizeof(rom_map_s));
					locator->number_of_entries = hton16(rom_map_entries);
					locator->magic = hton32(MAP_MAGIC);
					rom_map = (rom_map_s*)&image[ntoh32(locator->absolute_location)];
					if (f2a_verbose)
						print("Rom map table is at address 0x%x\nRom map locator is at address 0x%x (end of loader)\n",
							GBA_ROM + ntoh32(locator->absolute_location),
							GBA_ROM + rom_map_locator);
				}
				else
				{
					int oldfilesize = roundedfilesize;
					roundedfilesize++;
					adjust_rom_size(&roundedfilesize);
					// pad chunk with last loader byte
					memset(image + oldfilesize, last, roundedfilesize - oldfilesize);
					print("Increasing loader size to fit rom map by 0x%x/%i bytes\n", roundedfilesize - oldfilesize, roundedfilesize - oldfilesize);
				}
			} while (rom_map_entries < MAP_MINIMUM_ENTRIES);
			if (f2a_verbose)
				print("\n");

			print("Loader %s is:\n", loader.name);
			whole_loader_size = roundedfilesize;
		}
		else
		{
			print("File %s at address 0x%x is:\n", files[index], GBA_ROM + loadedsize);

			roundedfilesize = st.st_size;
			adjust_rom_size(&roundedfilesize);
			
			if (loadedsize + st.st_size > f2a_max_rom_size)
			{
				printerr("! WARNING ! Cart size exceeded, truncating file (size 0x%x -> 0x%x bytes) !\n", (int)st.st_size, f2a_max_rom_size - loadedsize);
				if ((st.st_size = f2a_max_rom_size - loadedsize) == 0)
					break;
				roundedfilesize = st.st_size;
				adjust_rom_size(&roundedfilesize);
			}

			// load ROM into image file
			if ((ROMf = fopen(files[index], "rb")) == NULL)
			{
				if (ferror(ROMf))
					printerrno("fopen(%s)", files[index]);
				else
					printerr("Could not open file %s for reading\n", files[index]);
				return -1;
			}
			if (fread(&image[loadedsize], st.st_size, 1, ROMf) != 1)
			{
				if (ferror(ROMf))
					printerrno("fread(%s)", files[index]);
				else
					printerr("Could not read file %s\n", files[index]);
				return -1;
			}
			fclose(ROMf);
			
			// try to reduce rom size
			if (f2a_reduce_allowed || f2a_reduce_forced)
			{
				int i, min_limit;
				// last bytes are probably padding
				char last = image[loadedsize + st.st_size - 1];

				// f2a_reduce_forced tries to mad-pad everything
				// !f2a_reduce_forced tries to fit loader only
				min_limit = f2a_reduce_forced? 0: MAX(0, (st.st_size - whole_loader_size - 1));
				// truncate as long as there is padding
				for (i = st.st_size - 2; i >= min_limit && image[loadedsize + i] == last; i--);
				roundedfilesize = i + 2;
				adjust_rom_size(&roundedfilesize);

				if (roundedfilesize < st.st_size)
				{
					print("(reducing ROM size from %ikB to %ikB", 
						((int)st.st_size + 1023) >> 10, 
						(roundedfilesize + 1023) >> 10);
					if (   (reduce += ((int)st.st_size - roundedfilesize)) > whole_loader_size
					    && !f2a_reduce_forced)
					{
						print(" - not trying to further reduce ROM size");
						f2a_reduce_allowed = 0;
					}
					print(")\n");
				}
			}
		}

		// check file against real ROM
		if (f2a_correct_header_allowed && index != -1)
			correct_header(&image[loadedsize], files[index]);
		display_map(&image[loadedsize]);
		
		// fill the rom locator
		if (locator)
		{
			if (index >= rom_map_entries)
			{
				printerr("Warning: not enough room in rom locator to stock further rom entries !\n");
				locator = NULL;
			}
			else
			{
				strncpy(rom_map[index].name, romname(&image[loadedsize]), MAP_NAMELEN);
				rom_map[index].offset = hton32(loadedsize);
				rom_map[index].size = hton32(roundedfilesize);
			}
		}
		
		if (!f2a_force_burn && has_same_data(loadedsize, roundedfilesize, image) >= 0)
		{
			print("No need to burn it!\n");
			// but it's time to burn the previous ones if they changed
			if (burnstart >= 0)
			{
				//XXXFIXME remove all this commentary when ACKed
				//	we don't burn anymore while loading, but record the location
				//	to burn into the chunks array because the map locator
				//	needs to be burned after all files are loaded, although it is
				//	located itself in the beginning of the rom
				//if (f2a_burn(GBA_ROM, image, burnstart, loadedsize - burnstart) < 0)
				//	return -1;
				
				// register chunk
				assert(chunks_used < chunks_number);
				chunks[chunks_used].offset = burnstart;
				chunks[chunks_used].size = loadedsize - burnstart;
				chunks_used++;
				
				burnstart = -1;
			}
		}
		else
		{
			print("Not the same !\n");
			// and we remember start address
			if (burnstart < 0)
				burnstart = loadedsize;
		}

		loadedsize += roundedfilesize;
	}

	// mark end of map
	if (locator && index < rom_map_entries)
	{
		rom_map[index].name[0] = 0;
		rom_map[index].offset = hton32(0);
		rom_map[index].size = hton32(0);
	}

	// burn the last chunk (XXXFIXMEremove this commentary)
	//if (burnstart >= 0 && f2a_burn(GBA_ROM, image, burnstart, loadedsize - burnstart))
	//	return -1;

	// register the last chunk
	if (burnstart >= 0)
	{
		assert(chunks_used < chunks_number);
		chunks[chunks_used].offset = burnstart;
		chunks[chunks_used].size = loadedsize - burnstart;
		chunks_used++;
	}
	
	print("\n");
	// now burn all the recorded chunks
	for (index = 0; index < chunks_used; index++)
		if (f2a_burn(GBA_ROM, image, chunks[index].offset, chunks[index].size) < 0)
			return -1;

        /* 
	 * Temporary write support for Ultra carts : we wipe out the descriptor and rely on
	 * CIZ to regenerate it
	 */
	if (cart_type == CART_TYPE_ULTRA)
	{
		printf("Wiping out content descriptor. You will need to launch CIZ to regenerate it.\n");
		memset((char*) &desc, 0x99, sizeof(desc));
		f2a_writemem((unsigned char*) &desc, F2AU_CD_BASE, 0, SIZE_64K, f2a_write_block_size);
	}
	
	print("Remaining size: %iMbits = %iMBytes = %iKBytes\n",
		(f2a_max_rom_size - loadedsize) / 1024 / 1024 * 8,
		(f2a_max_rom_size - loadedsize) / 1024 / 1024,
		(f2a_max_rom_size - loadedsize) / 1024);

	if (clean_rom)
	{
		//int cleanblocksize = f2a_write_block_size; // too big, unnecessary
		//int cleanblocksize = HEADERSIZE_B; // don't work (?)
		int cleanblocksize = f2a_rom_block_size; // 32Kb

		// we write cleanblocksize bytes every f2a_write_block_size, I
		// suppose that each write cleans f2a_write_block_size
		// (>=cleanblocksize) bytes. to be verified...
		char empty[cleanblocksize];

		int dummy = 0;
		print("\nCleaning remaining ROM...\n");
		memset(empty, 0xff, cleanblocksize);
		burnstart = loadedsize + f2a_write_block_size - 1;
		adjust_burn_addresses(&burnstart, &dummy);
		for (; burnstart < f2a_max_rom_size; burnstart += f2a_write_block_size)
		{
			print("Cleaning from 0x%x to 0x%x\n", GBA_ROM + burnstart, 
				GBA_ROM + burnstart + cleanblocksize);
			f2a_writemem(empty, GBA_ROM, burnstart, cleanblocksize, cleanblocksize);
			//break; no break: need to clear everything left
		}
		print("\n");
	}
	
	free(image);
	return 0;
}

int f2a_load_cart_map (void)
{
	char			small_rom [LOADUNIT];
	char*			rom	= NULL;
	rom_map_locator_s*	locator	= NULL;
	rom_map_s*		rom_map	= NULL;
	int			offset, i;
	
	if (f2a_sim > 1)
		return 0;
	
	// find locator
	print("Searching for locator...\n");
	for (offset = 0; !locator && offset < f2a_max_rom_size; offset += f2a_rom_block_size)
	{
		if (f2a_readmem(small_rom, GBA_ROM + offset + f2a_rom_block_size - LOADUNIT, LOADUNIT) == -1)
			return -1;
		locator = (rom_map_locator_s*)&small_rom[LOADUNIT - sizeof(rom_map_locator_s)];
		if (ntoh32(locator->magic) == MAP_MAGIC)
		{
			int rom_map_offset;
			int rom_map_size;

			if (f2a_verbose)
			{
				print("Rom map locator found !\n");
				print("Rom map offset=%x size=%i roms\n", GBA_ROM + ntoh32(locator->absolute_location), ntoh16(locator->number_of_entries));
			}

			// reload in case our small_rom is not big enough to get all entries
			rom_map_offset = GBA_ROM + ntoh32(locator->absolute_location);
			rom_map_size = ntoh16(locator->number_of_entries) * sizeof(rom_map_s);
			adjust_load_addresses(&rom_map_offset, &rom_map_size);
			if ((rom = (char*)malloc(rom_map_size)) == NULL)
			{
				printerrno("Cannot allocate %i bytes to load rom map: malloc:");
				return -1;
			}
			if (f2a_readmem(rom, rom_map_offset, rom_map_size) == -1)
				return -1;
			rom_map = (rom_map_s*)&rom[ntoh32(locator->absolute_location) - (rom_map_offset - GBA_ROM)];

			print("Cart rom map:\n");
			print("Loader / First rom size is %.4gMBits (including map), %.4gMbits (excluding map).\n",
				(offset + f2a_rom_block_size) * 8.0 / 1024 / 1024,
				ntoh32(locator->absolute_location) * 8.0 / 1024 / 1024);
			for (i = 0; i < ntoh16(locator->number_of_entries) && ntoh32(rom_map[i].size) > 0; i++)
				print("rom '%s' - offset=0x%x/%.4gMbits size=0x%x/%.4gMbits\n",
					rom_map[i].name,
					ntoh32(rom_map[i].offset),
					ntoh32(rom_map[i].offset) * 8.0 / 1024 / 1024,
					ntoh32(rom_map[i].size),
					ntoh32(rom_map[i].size) * 8.0 / 1024 / 1024);
			if (i > 0)
				print("rom map stops at %.4gMBits.\n", 1.0 * (ntoh32(rom_map[i - 1].offset) + ntoh32(rom_map[i - 1].size)) * 8.0 / 1024 / 1024);
			
			free(rom);
			return 0;
		}
		else
			locator = NULL;
	}
	printerr("Could not find rom map locator (it will be generated next time a rom is burnt).\n");
	return -1;
}

void adjust_burn_addresses (int* offset, int* size)
{
	// adjust offset so that it's a multiple of block
	// adjust size so that it's a multiple of block
	// (this is needed if we want to rewrite a rom between two others)
	*size += *offset;
	*offset &= ~(f2a_write_block_size - 1);
	*size -= *offset;
	if (*size & (f2a_write_block_size - 1))
		*size += f2a_write_block_size;
	*size &= ~(f2a_write_block_size - 1);
}

void adjust_load_addresses (int* offset, int* size)
{
	// adjust offset so that it's a multiple of block
	// adjust size so that it's a multiple of block
	*size += *offset;
	*offset &= ~(LOADUNIT - 1);
	*size -= *offset;
	if (*size & (LOADUNIT - 1))
		*size += LOADUNIT;
	*size &= ~(LOADUNIT - 1);
}

void adjust_write_size (int* size)
{
	if (*size & (f2a_write_block_size - 1))
		*size += f2a_write_block_size;
	*size &= ~(f2a_write_block_size - 1);
}

void adjust_rom_size (int* size)
{
	if (*size & (f2a_rom_block_size - 1))
		*size += f2a_rom_block_size;
	*size &= ~(f2a_rom_block_size - 1);
}
