/* 
 * Based on f2a by Ulrich Hecht <uli@emulinks.de>
 * if2a patch by D. Gauchard <deyv@free.fr>
 * F2A Ultra support by Vincent Rubiolo <vincent.rubiolo@free.fr>
 * Firmware loading debug by Julien Janier <julien@janier.org>
 * Licensed under the terms of the GNU Public License version 2
 */

#include "libf2a.h"
#include "f2ausb.h"
#include "f2aio.h"
#include "f2autils.h"

#if !_WIN32
#include <unistd.h>
#include <sys/utsname.h>	// uname()
#include <sys/wait.h>		// sleep
#endif

#define EZUSB2131_MODULE "/proc/ezusb/dev0"

struct usb_bus *busses;
struct usb_bus *bus;
struct usb_device *dev;
struct usb_device *f2a = NULL;
usb_dev_handle *f2ahandle = NULL;
int f2a_usb_timeout = USB_TIMEOUT;

/* This function tests against the kernel version to know whether a USB reset 
 * and usb_set_configuration is needed (or unwanted).
 * If the machine runs linux-2.6, we do need to reset, if it is 2.4, we must not.
 * Also under 2.4, usb_set_configuration won't work whereas it is required
 * under win32.
 * 
 * We pass a verbosity parameter instead of relying on the global verbosity 
 * level because otherwise the kernel info would be printed each time we are 
 * called.
 */
int f2a_linux24 (int verbosity)
{
#if !_WIN32 && !defined(__CYGWIN32__)

	struct utsname utsname;
	int version, patchlevel;
	
	if (uname(&utsname) == -1)
	{
		printerrno("utsname");
		return 0;
	}

	version = atoi(strtok(utsname.release, "."));
	patchlevel = atoi(strtok(NULL, "."));

	if (verbosity == 1)
	{
		print("System informations:\n");
		print("\tSystem name = %s\n", utsname.sysname);
		print("\tNode name = %s\n", utsname.nodename);
		print("\tRelease = %s\n", utsname.release);
		print("\tVersion = %s\n", utsname.version);
		print("\tMachine = %s\n", utsname.machine);
	}

	if (!strcmp(utsname.sysname, "Linux"))
	{
		if (verbosity == 1)
			print("\tRunning on Linux %i.%i\n", version, patchlevel);
	}
	else
		return 1;
		
	return version == 2 && patchlevel == 4;
#else
	return 0;

#endif
}


int f2a_fxload (char* bus, char* dev)
{
#if _WIN32 || defined(__CYGWIN32__)

	return -1;
	
#else // !WIN32
#if INTERNAL_FXLOAD

	int f, status;
	char ez[22];
	
	/* Trying to use fxload code (2.4.19 - 2.6) */
	snprintf(ez, 22, "/proc/bus/usb/%s/%s", bus, dev);
	if (f2a_verbose)
		print("Uploading EZ-USB firmware in %s\n", ez);
	if ((f = open(ez, O_WRONLY)) == -1)
	{
		printerr("Trying to use internal fxload (linux-2.4.19+): %s: %s\n", ez, strerror(errno));
		return -1;
	}
	status = ezusb_load_ram(f);
	close(f);
	if (status != 0)
	{
		printerr("Error while sending firmware\n");
		return -1;
	}
	return 0;

#else // !INTERNAL_FXLOAD

	int fd;
	int result;
	char cmd[128], file[128];

	snprintf(file, 128, "/tmp/if2a-f2afirm-%i.hex", getpid());
	if ((fd = open(file, O_CREAT | O_EXCL | O_WRONLY, 0600)) == -1)
	{
		printerrno(file);
		return -1;
	}
	if (write(fd, firmware.data, firmware.size) != firmware.size)
	{
		printerrno("write");
		close(fd);
		unlink(file);
		return -1;
	}
	if (close(fd) != 0)
	{
		printerrno("close");
		unlink(file);
		return -1;
	}

	//FIXME use forkexec instead?
	//sprint(cmd, "/sbin/fxload -D /proc/bus/usb/%s/%s -I %s -t an21 -vvv", bus, dev, file);
	snprintf(cmd, 128, "/sbin/fxload -D /proc/bus/usb/%s/%s -I %s -t an21", bus, dev, file);
	print("Executing '%s':\n", cmd);
	result = system(cmd);
	print("result = %i - ", result);

	if (result == 0)
		print("FXLOAD Success\n");
	else
		print("FXLOAD Error?\n");

	if (unlink(file) != 0)
		printerrno("unlink");

	return result ? -1 : 0;

#endif // !INTERNAL_FXLOAD
#endif // !def _WIN32
}

int f2a_ezusb2131_load (int f)
{
	int w, tot;
	for (tot = 0; tot < firmware.size; tot += w)
	{
		w = write(f, firmware.data + tot, firmware.size - tot);
		if (w == -1)
		{
			printerrno("write on "EZUSB2131_MODULE);
			return -1;
		}
		
		if (f2a_verbose > 1)
		{
			print(EZUSB2131_MODULE": wrote %i bytes [%i->%i / %i]\r", w, tot, tot+w, firmware.size);
		}
	}
	printf("\n");
	return 0;
}

int f2a_usb_connect_root(void)
{
	int firmware_loaded=0;
	int linux_24 = f2a_linux24(f2a_verbose > 1);
	//usb_set_debug(1);
	usb_init();
	usb_find_busses();

find_f2a:
	usb_find_devices();
	busses=usb_busses;
	
	f2a = NULL;

	for (bus = busses; !f2a && bus; bus = bus->next)
		for (dev = bus->devices; !f2a && dev; dev = dev->next)
			if ((dev->descriptor.idVendor == 0x547) && (dev->descriptor.idProduct == 0x1002))
			{
				f2a = dev;
				break;
			}
			else if ((dev->descriptor.idVendor == 0x547) && (dev->descriptor.idProduct == 0x2131) && !firmware_loaded)
			{
				int f;
				
				/* Trying first to use ezusb2131 module (linux 2.4) */
				if ((f = open(EZUSB2131_MODULE, O_WRONLY)) != -1)
				{
					if (f2a_verbose)
						print("Uploading EZ-USB firmware in "EZUSB2131_MODULE"\n");
					int result = f2a_ezusb2131_load(f);
					close(f);
					if (result != 0)
						return -1;
				}
				else
				{
					int ezusb2131_errno = errno;
					if (f2a_fxload(bus->dirname, dev->filename) != 0)
					{
						printerr("Trying to use ezusb2131 module (linux-2.4): "EZUSB2131_MODULE": %s\n", strerror(ezusb2131_errno));
						return -1;
					}
				}
			
				firmware_loaded = 1;
				if (f2a_verbose > 1)
					printf("EZ-USB renumerating. Please wait\n");
				sleep(2);	// give the EZUSB some time to renumerate
				goto find_f2a;
			}

	if (f2a == NULL)
	{
		printerr("Error: couldn't find f2a attached to usb\n");
		return -1;
	}

	// Open the USB device and get handle
	if ((f2ahandle = usb_open(f2a)) == NULL)
	{
		printerr("usb_open: %s\n", usb_strerror());
		return -1;
	}
	
	// Set configuration (needed for win32, unwanted in linux-2.4, no matter for linux-2.6)
	if (!linux_24 && usb_set_configuration(f2ahandle, 1) < 0)
	{
		printerr("usb_set_configuration: %s\n", usb_strerror());
		return -1;
	}

	// Reset if needed (yes for 2.6 kernels, no for 2.4 kernel, no matter for win32?)
	if (!linux_24 && usb_reset(f2ahandle) < 0)
	{
		printerr("usb_reset: %s\n", usb_strerror());
		return -1;
	}

	/* Be clean : claim interface. F2A linker has one interface (#0) and 
	 * 3 endpoints : 0x0 (control), 0x04 (to write into f2a), 0x83 (to 
	 * read)
	 */
	if (usb_claim_interface(f2ahandle, 0x0) < 0)
	{
		printerr("usb_claim_interface: %s\n", usb_strerror());
		return -1;
	}

	return 0;
}

int f2a_usb_connect (void)
{
	int result = f2a_usb_connect_root();
#if !_WIN32
	seteuid(getuid());
#endif
	return result;
}

void f2a_usb_disconnect (void)
{
	/* F2A linker device has one interface (0x0)
	 * and 3 endpoints : 0x0 - control -, 0x4 - write -,
	 * 0x83 - read -
	 */
	if (usb_release_interface(f2ahandle, 0x0) < 0)
		printerr("usb_release: %s\n", usb_strerror());

	if (usb_close(f2ahandle) < 0)
		printerr("usb_close: %s\n", usb_strerror());
}

void f2a_usb_exit (int status)
{
	if (f2ahandle)
		f2a_usb_disconnect();
	exit(status);
}

int f2a_usb_read (char *buffer, int buffer_size)
{
	int result;

	result = usb_bulk_read(f2ahandle, 0x83, buffer, buffer_size, f2a_usb_timeout);
	if (result != buffer_size)
	{
		printerr("usb_bulk_read: %s\n", usb_strerror());
		return -1;
	}
	return result;
}

int f2a_usb_write(char *buffer, int buffer_size)
{
	int result;

	result = usb_bulk_write(f2ahandle, 0x4, buffer, buffer_size, f2a_usb_timeout);
	if (result != buffer_size)
	{
		printerr("usb_bulk_write: %s\n", usb_strerror());
		return -1;
	}
	return result;
}

int f2a_usb_write_by_block (char* buffer, int buffer_size, int block_size)
{
	int sent, result;
	
	sent = 0;
	while (sent < buffer_size)
	{
		result = f2a_usb_write(buffer + sent, buffer_size - sent > block_size? block_size: buffer_size - sent);
		if (result == -1)
			return -1;
		sent += result;
	}
	return sent;
}
