/*
 * part of fxload (http://linux-hotplug.sourceforge.net):
 * Copyright (c) 2001 Stephen Williams (steve@icarus.com)
 * Copyright (c) 2001-2002 David Brownell (dbrownell@users.sourceforge.net)
 * Arrangement for Flash2Advance USB cable - Julien Janier (julien@janier.org)
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "libf2a.h"
#include "f2aio.h"

#if INTERNAL_FXLOAD

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>

#include <linux/types.h>
typedef unsigned long kernel_ulong_t;
#if NEED_LE16
typedef __u16 __le16;
#endif

#include <sys/ioctl.h>

#include <linux/version.h>
#include <linux/usb.h>
#include <linux/usbdevice_fs.h>

#define RETRY_LIMIT	5

#define RW_INTERNAL     0xA0	/* hardware implements this one */
#define RW_EEPROM       0xA2
#define RW_MEMORY       0xA3
#define GET_EEPROM_SIZE 0xA5

typedef enum
{
	_undef = 0,
	internal_only,		/* hardware first-stage loader */
	skip_internal,		/* first phase, second-stage loader */
	skip_external		/* second phase, second-stage loader */
} ram_mode;


struct ram_poke_context
{
	int device;
	ram_mode mode;
	unsigned total, count;
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,3)
/*
 * in 2.5, "struct usbdevfs_ctrltransfer" fields were renamed
 * to match the USB spec
*/
#       define bRequestType     requesttype
#       define bRequest         request
#       define wValue           value
#       define wIndex           index
#       define wLength          length
#endif


static inline int ctrl_msg(int device, unsigned char requestType, unsigned char request, unsigned short value, unsigned short index, unsigned char *data, size_t length)
{
	struct usbdevfs_ctrltransfer ctrl;

	if (length > USHRT_MAX)
	{
		fputs("length too big", stderr);
		return -EINVAL;
	}

	/* 8 bytes SETUP */
	ctrl.bRequestType = requestType;
	ctrl.bRequest = request;
	ctrl.wValue = value;
	ctrl.wLength = (unsigned short)length;
	ctrl.wIndex = index;

	/* "length" bytes DATA */
	ctrl.data = data;

	ctrl.timeout = 10000;

	return ioctl(device, USBDEVFS_CONTROL, &ctrl);
}


/*
 * Modifies the CPUCS register to stop or reset the CPU.
 * Returns false on error.
 */
static int ezusb_cpucs(int device, unsigned short addr, int doRun)
{
	int status;
	unsigned char data = doRun ? 0 : 1;

	if (f2a_verbose)
		printerr("%s\n", data ? "stop CPU" : "reset CPU");
	status = ctrl_msg(device, USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE, RW_INTERNAL, addr, 0, &data, 1);
	if (status != 1)
	{
		char *mesg = "can't modify CPUCS";
		if (status < 0)
			printerrno(mesg);
		else
			printerr("%s\n", mesg);
		return 0;
	}
	else
		return 1;
}

/*
 * Issues the specified vendor-specific write request.
 */
static int ezusb_write(int device, char *label, unsigned char opcode, unsigned short addr, const unsigned char *data, size_t len)
{
	int status;

	if (f2a_verbose)
		printerr("%s, addr 0x%04x len %4d (0x%04x)\n", label, addr, len, len);
	status = ctrl_msg(device, USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE, opcode, addr, 0, (unsigned char *)data, len);
	if (status != len)
	{
		if (status < 0)
			printerrno(label);
		else
			printerr("%s ==> %d\n", label, status);
	}
	return status;
}

static int poke(void *context, unsigned short addr, int external, const unsigned char *data, size_t len)
{
	struct ram_poke_context *ctx = context;
	int rc;
	unsigned retry = 0;

	switch (ctx->mode)
	{
	case internal_only:	/* CPU should be stopped */
		if (external)
		{
			printerr("can't write %d bytes external memory at 0x%04x\n", len, addr);
			return -EINVAL;
		}
		break;
	case skip_internal:	/* CPU must be running */
		if (!external)
		{
			if (f2a_verbose >= 2)
				printerr("SKIP on-chip RAM, %d bytes at 0x%04x\n", len, addr);
			return 0;
		}
		break;
	case skip_external:	/* CPU should be stopped */
		if (external)
		{
			if (f2a_verbose >= 2)
				printerr("SKIP external RAM, %d bytes at 0x%04x\n", len, addr);
			return 0;
		}
		break;
	default:
		printerr("bug\n");
		return -EDOM;
	}

	ctx->total += len;
	ctx->count++;

	/* Retry this till we get a real error. Control messages are not
	 * NAKed (just dropped) so time out means is a real problem.
	 */
	while ((rc = ezusb_write(ctx->device, external ? "write external" : "write on-chip", external ? RW_MEMORY : RW_INTERNAL, addr, data, len)) < 0 && retry < RETRY_LIMIT)
	{
		if (errno != ETIMEDOUT)
			break;
		retry += 1;
	}
	return (rc < 0) ? -errno : 0;
}

int is_external(unsigned short addr, size_t len)
{
	/* with 8KB RAM, 0x0000-0x1b3f can be written
	 * we can't tell if it's a 4KB device here
	 */
	if (addr <= 0x1b3f)
		return ((addr + len) > 0x1b40);

	/* there may be more RAM; unclear if we can write it.
	 * some bulk buffers may be unused, 0x1b3f-0x1f3f
	 * firmware can set ISODISAB for 2KB at 0x2000-0x27ff
	 */
	return 1;
}

unsigned char *skip_line(unsigned char *buffer)
{
	while ((*buffer != '\n' || *buffer != '\r') && *buffer != '\0')
		buffer++;
	if (*buffer == '\r' || *buffer == '\n')
		buffer++;
	return buffer;
}

int parse_ihex(unsigned char *image, void *context)
{

	unsigned char data[1023];
	unsigned short data_addr = 0;
	size_t data_len = 0;
	int rc = -1;
	int first_line = 1;
	int external = 0;

	char *buf, *cp;

	buf = image;
	for (;;)
	{
		char tmp, type;
		size_t len;
		unsigned idx, off;


		/* test if end of buffer reached */
		if (buf[0] == '\0')
		{
			printerr("not an ihex record: %s", buf);
			return -2;
		}

		/* skip cariage return */
		while (buf[0] == '\r' || buf[0] == '\n')
			buf++;

		/* EXTENSION: "# comment-till-end-of-line", for copyrights etc */
		if (buf[0] == '#')
		{
			buf = skip_line(buf);
			continue;
		}

		if (buf[0] != ':')
		{
			printerr("not an ihex record: %s\n", buf);

			printerr("{%x}\n", *buf);

			return -2;
		}

		/* Read the length field (up to 16 bytes) */
		tmp = buf[3];
		buf[3] = 0;
		len = strtoul(buf + 1, 0, 16);
		buf[3] = tmp;

		/* Read the target offset (address up to 64KB) */
		tmp = buf[7];
		buf[7] = 0;
		off = strtoul(buf + 3, 0, 16);
		buf[7] = tmp;

		/* Initialize data_addr */
		if (first_line)
		{
			data_addr = off;
			first_line = 0;
		}

		/* Read the record type */
		tmp = buf[9];
		buf[9] = 0;
		type = strtoul(buf + 7, 0, 16);
		buf[9] = tmp;

		/* printerr("%d - %d - %d\n", off, len, type); */

		/* If this is an EOF record, then make it so. */
		if (type == 1)
		{
			if (f2a_verbose > 2)
				printerr("Read EOF entry\n");
			break;
		}

		if (type != 0)
		{
			printerr("unsupported record type: %u\n", type);
			return -3;
		}

		/* flush the saved data if it's not contiguous,
		 * or when we've buffered as much as we can.
		 */
		if (data_len != 0 && (off != (data_addr + data_len)
				      // || !merge
				      || (data_len + len) > sizeof data))
		{
			external = is_external(data_addr, data_len);
			rc = poke(context, data_addr, external, data, data_len);

			if (rc < 0)
				return -1;
			if (f2a_verbose > 2)
				printerr("Write %d bytes to ezusb\n", data_len);
			data_addr = off;
			data_len = 0;
		}

		/* append to saved data, flush later */
		for (idx = 0, cp = buf + 9; idx < len; idx += 1, cp += 2)
		{
			tmp = cp[2];
			cp[2] = 0;
			data[data_len + idx] = strtoul(cp, 0, 16);
			cp[2] = tmp;
		}
		buf += len * 2 + 11;
		data_len += len;
	}


	/* flush any data remaining */
	if (data_len != 0)
	{
		external = is_external(data_addr, data_len);
		rc = poke(context, data_addr, external, data, data_len);
		if (rc < 0)
			return -1;
		if (f2a_verbose > 2)
			printerr("Write %d bytes to ezusb\n", data_len);
	}
	return 0;
}

int ezusb_load_ram(int fd)
{
	unsigned short cpucs_addr;
	struct ram_poke_context ctx;
	int status;

	cpucs_addr = 0x7f92;

	ctx.mode = internal_only;

	/* don't let CPU run while we overwrite its code/data */
	if (!ezusb_cpucs(fd, cpucs_addr, 0))
		return -1;

	/* scan the image, first (maybe only) time */
	ctx.device = fd;
	ctx.total = ctx.count = 0;
	status = parse_ihex(firmware.data, &ctx);
	if (status < 0)
	{
		printerr("unable to download firmware");
		return status;
	}

	if (f2a_verbose)
		printerr("... WROTE: %d bytes, %d segments, avg %d\n", ctx.total, ctx.count, ctx.total / ctx.count);

	/* now reset the CPU so it runs what we just downloaded */
	if (!ezusb_cpucs(fd, cpucs_addr, 1))
		return -1;

	return 0;
}

#endif // INTERNAL_FXLOAD
