/* 
 * Based on f2a by Ulrich Hecht <uli@emulinks.de>
 * if2a patch by D. Gauchard <deyv@free.fr>
 * F2A Ultra support by Vincent Rubiolo <vincent.rubiolo@free.fr>
 * Licensed under the terms of the GNU Public License version 2
 */

// Main f2a header

#ifndef __F2A_H__
#define __F2A_H__

#define VERSION			"0.93-pre0"

#if _WIN32
	#include <windows.h>
	#define sleep(n) Sleep(1000*(n))
	typedef unsigned int u_int32_t; // ANSI-C does not define u_int32_t
#else
	#include <sys/types.h> // u_int32_t
#endif

///////////////////////////////////////////////
//FIXME: use more explicit names (f2a_* F2A_*):
///////////////////////////////////////////////

//////////////////////////////////////
// general defines

#define GBA_HEADNAME		"GBAROM-"
#define GBA_EXTNAME		".gba"

#define MAX(x,y)		((x)>(y)? (x): (y))
#define MIN(x,y)		((x)<(y)? (x): (y))
#define ASCII(x)		(((x) >= 32 && isascii(x))? (x): '.')

#define HEADERSIZE_B		0x9C		// 156 bytes
#define SIZE_64K		65536
#define MAXBURNCHUNK		(8 << 20)	// maximum burning size at once in bytes - 8MB (do not raise!)
#define LOADUNIT		1024		// minimum read/write data size on cart

enum mode_e
{
	MODE_WRITE,
	MODE_READ,
	MODE_UNDEF,
};

enum read_type_e
{
	READ_ONCE,
	READ_MANY,				// several GBA memory areas into a file
};

#define GBA_EWRAM		0x02000000	// ram address for multiboot files
#define GBA_VRAM		0x06000000	// Video RAM for loader background image
#define GBA_OAM			0x07000000	// OAM (Object Attribute Memory?)
#define	GBA_ROM			0x08000000
#define GBA_SRAM		0x0e000000	// cart SRAM

#define	F2AU_DH_BASE		(GBA_SRAM + 0x370000)
#define F2AU_SVD_BASE		(GBA_SRAM + 0x020000)
#define F2AU_CD_BASE		(GBA_SRAM + 0x0f0000)

#define MAGIC_NUMBER		0xa46e5b91	// needs to be properly set for almost all F2A commands

enum cart_type_e
{
	CART_TYPE_TURBO,
	CART_TYPE_PRO,
	CART_TYPE_ULTRA,
	CART_TYPE_UNDEF,
};
#define cart_type_str(type) ((char*)(type == CART_TYPE_TURBO? "Turbo": type == CART_TYPE_PRO? "Pro": type == CART_TYPE_ULTRA? "Ultra": "(Not detected)"))

//////////////////////////////////////
// borrowed from binware.h
#ifndef __BINWARE_H__
#define __BINWARE_H__

typedef struct
{
	unsigned char*	name;
	int		size;
	unsigned char*	data;
} binware_s;

extern binware_s binware_loader_pro [];
extern binware_s binware_loader_ultra [];
extern binware_s binware_firmware [];
extern binware_s binware_multiboot [];
extern binware_s binware_splash [];
#endif // __BINWARE_H__

//////////////////////////////////////
// libf2a internal variables (rename us, should use accessors)
extern int		f2a_verbose;		// 0, 1, 2...
extern binware_s	loader;			// ROM loader binaries
extern binware_s	firmware;		// Linker firmware binaries
extern binware_s	multiboot;		// F2A multiboot binaries
extern binware_s	splash;			// F2A splash binaries


//////////////////////////////////////
// library general functions

// borrowed from f2autils.h

void		f2a_init		(void);
int		f2a_connect		(void);
void		f2a_exit		(int status);
int		select_loader_pro	(char* name);
int		select_loader_ultra	(char* name);
int		select_firmware		(char* name);
int		select_multiboot	(char* name);
int		select_splash		(char* name);
int		buffer_from_file	(const char* filename, char* buffer, int size_to_check);
int		buffer_to_file		(const char* filename, char* buffer, int size_to_write);

// borrowed from f2acarts.h

extern int	f2a_thorough_comp;
extern int	f2a_use_loader;
extern int	f2a_correct_header_allowed;
extern int	f2a_force_burn;
extern int	f2a_reduce_forced;
extern int	f2a_reduce_allowed;

int		select_loader           (enum cart_type_e cart_type, char* loader_file);
int		convsize                (char* size);
int		convf2ablock            (char* block, int* offset, int* size);
int		display_cart_map        (void);
int		f2apro_loadandwrite_sram(char* file, int offset, int size);
int		auto_loadandburn_rom    (enum cart_type_e cart_type, int clean_rom, int numfiles, char** files);
void		auto_readandsave_rom    (int numfiles, char** files);
int		f2a_load_cart_map	(void);

// borrowed from f2aio.h

extern int	f2a_rom_block_size_log2;	// log2(f2a_rom_block_size) - roms have to have a size multiple of f2a_rom_block_size
extern int	f2a_write_block_size_log2;	// log2(f2a_write_block_size) - minimum block size to write on cart
extern int	f2a_cartsize_mb; 		// cart size in Mbits. Now set by f2a_get_type()
extern int	f2a_sim;			// 0: normal, 1: read, no write, >1: no read, no write
// wrappers
#define		f2a_max_rom_size		((f2a_cartsize_mb) * 1024 * 1024 / 8)
#define		f2a_rom_block_size		(1<<f2a_rom_block_size_log2)
#define		f2a_write_block_size		(1<<f2a_write_block_size_log2)

enum cart_type_e f2a_get_type			(int* size_mbits, int* f2a_write_block_size_log2, int* f2a_rom_block_size_log2);
int		f2a_boot 			(binware_s* multiboot, binware_s* splash, int usb_timeout_in_seconds);
int		f2a_info 			(void);
int		f2a_multiboot 			(char* fileName);
int		f2a_readmem			(unsigned char* data, int address, int size); //fixme: remove me from here
int		f2a_readmemtofile		(char* file, int address, int size, enum read_type_e read_type);

//////////////////////////////////////
// f2a ultra definitions
// borrowed from f2aultra.h

#ifndef __F2A_ULTRA_TYPES
#define __F2A_ULTRA_TYPES

// Each ROM has such an entry in the content descriptor
typedef struct rom_entry_s
{
	u_int32_t     list_id;			// ROM nb in the release lists (shows up in CIZ)
	u_int32_t     game_id;			// GameID displayed by CIZ. CRC-based
	unsigned char rom_name[12];		// ASCII ROM name
	unsigned char unknown1[4];
	u_int32_t     rom_size;			// ROM size in bytes
	u_int32_t     _padding1[2];
	unsigned char unknown2[2];
	unsigned char ciz_idx;			// index in CIZ list. Index starts at 0
	unsigned char unknown3;
	u_int32_t     list_crc;			// ROM CRC from release lists
	unsigned char rom_lg[2];		// ROM language (shows up in CIZ)
	unsigned char rom_saver[2];		// ROM saver type (shows up in CIZ)
	unsigned char unknown5[6];
	unsigned char _padding2[10];
	
} __attribute__ ((packed)) rom_entry;

// Defines cart contents - see PROTOCOL for more details
typedef struct content_desc_s
{
	u_int32_t	_fixed1[896];		// zeroes ?
	u_int32_t	lp_list_id;		// last played ROM ID from release lists
	u_int32_t	lp_game_id;		// last played ROM GameID
	u_int32_t	lp_idx;			// last played ROM index in CIZ menu
	u_int32_t	_fixed2[125];		// zeroes ?
	u_int32_t	unknown;		// probably CRC on descriptor
	u_int32_t	_padding1[5];
	u_int32_t	fixed3;			// 0x8000021e
	u_int32_t	fixed4;			// 0x01000000
	u_int32_t	_padding2[2];
	unsigned char	fixed5;			// zeroes ?
	unsigned char	nb_visible;		// number of visible roms in CIZ menu
	unsigned char	fixed6[214];
	rom_entry	roms[100];		// ROM entries (see above)
	u_int32_t	_padding5[13696];

} __attribute__ ((packed)) content_desc;

#endif // __F2A_ULTRA_TYPES

int		f2au_SVD_to_file        (char* filename);
int		f2au_SVD_from_file      (char* file);
int		f2au_DH_to_file         (char* file);
int		f2au_DH_from_file       (const char* file);
content_desc*	f2au_CD_read		(void);
int		f2au_CD_check           (content_desc* desc);
int		f2au_CD_print           (content_desc* descriptor);
int		f2au_GameID_gen         (const char* filename, int* game_id);
int		f2au_loadandwrite_sram	(char* file, int offset, int size);

//////////////////////////////////////
// print functions called by libf2a
// * print, printerr and printerrno have exactly the same syntax as printf()
//   print is used for regular message
//   printerr for error message
//   printerrno for error messages including errno value
//   (ex: perror("text") becomes printerrno("text\n") but one can do printerrno("open(%s)", filename))
// * useful: print & printerr may be redefined so to catch the text and do whatever nedded with it,
//           printerrno() calls printerr()
// * default: print calls printf, printerr calls fprintf(stderr)

typedef void (*print_f) (const char*, ...);
extern print_f	print;
extern print_f	printerr;
void		printerrno (const char* format, ...);

#endif // __F2A_H__
